# =====================================
# generator=datazen
# version=3.1.3
# hash=50538057667f048e4e1c2e1232e77e1b
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "An interface generator for distributed computing."
PKG_NAME = "ifgen"
VERSION = "2.4.0"
